from .. import ws
from ..codes import MSG_FIELD, GRID_EVENTS

from .socket_handler import SocketHandler
from .network import *
import json

socket_handler = SocketHandler()

routes = {
    GRID_EVENTS.JOIN: register_node,
    GRID_EVENTS.MONITOR_ANSWER: update_node,
    GRID_EVENTS.FORWARD: forward,
}


def route_request(message, socket):
    global routes

    message = json.loads(message)
    print("Message: ", message)
    if message and message.get(MSG_FIELD.TYPE, None) in routes.keys():
        return routes[message[MSG_FIELD.TYPE]](message, socket)
    else:
        return {"status": "error", "message": "Invalid request format!"}


@ws.route("/")
def socket_api(socket):
    while not socket.closed:
        message = socket.receive()
        if not message:
            continue
        else:
            response = route_request(message, socket)
            if response:
                socket.send(json.dumps(response))

    worker_id = socket_handler.remove(socket)



from .. import http
from ..events import SocketHandler

from flask import Response
import json

socket_handler = SocketHandler()


@http.route("/datasets", methods=["GET"])
def get_datasets():
    try:
        response = {}

        for node_id, node in socket_handler.nodes:
            response[node_id] = node.hosted_datasets

        response_body = json.dumps(response)
        return Response(response_body, status=200, mimetype="application/json")
    except Exception:
        return Response({}, status=500, mimetype="application/json")



from .. import http
from ..events.socket_handler import SocketHandler
from ..codes import MSG_FIELD
from flask import Response
import json

socket_handler = SocketHandler()

@http.route("/node/<id>")
def node_infos(id):
    try:
        worker = socket_handler.get(id)
        response = {}

        if worker:
            response["id"] = worker._id
            response["status"] = worker.status
            response["address"] = worker.address
            response["nodes"] = worker.connected_nodes
            response["datasets"] = worker.hosted_datasets
            response["models"] = list(worker.hosted_models.keys())
            response["cpu"] = worker.cpu_percent
            response["memory"] = worker.mem_usage

            response_body = json.dumps(response)
            return Response(response_body, status=200, mimetype="application/json")
        else:
            return Response(
                {
                    MSG_FIELD.STATUS: MSG_FIELD.ERROR,
                    MSG_FIELD.CONTENT: "Worker ID not found!",
                },
                status=404,
                mimetype="application/json",
            )
    except Exception:
        return Response({}, status=500, mimetype="application/json")



from gevent import pywsgi
from geventwebsocket.handler import WebSocketHandler
from flask import Flask, Blueprint
from flask_sockets import Sockets

ws = Blueprint(r"ws", __name__)
http = Blueprint(r"http", __name__)
DEFAULT_SECRET_KEY = "justasecretkeythatishouldputhere"
__version__ = "0.1.0"

def create_app(debug=False, secret_key=DEFAULT_SECRET_KEY):
    app = Flask(__name__)
    app.debug = debug
    app.config["SECRET_KEY"] = secret_key

    # Global socket handler
    sockets = Sockets(app)

    app.register_blueprint(http, url_prefix=r"/")
    sockets.register_blueprint(ws, url_prefix=r"/")

    return app

def raise_grid():
    app = create_app()
    server = pywsgi.WSGIServer(("", 5000), app, handler_class=WebSocketHandler)
    server.serve_forever()
    return app, server



"""
    All Gateway routes (REST API).
"""
from .. import http
from flask import render_template


@http.route("/", methods=["GET"])
def index():
    """ Main Page. """
    return "Open Grid Network"



class GRID_EVENTS:
    JOIN = "join"
    FORWARD = "forward"
    MONITOR_ANSWER = "monitor-answer"


class NODE_EVENTS:
    MONITOR = "monitor"
    WEBRTC_SCOPE = "create-webrtc-scope"
    WEBRTC_OFFER = "webrtc-offer"
    WEBRTC_ANSWER = "webrtc-answer"


class MSG_FIELD:
    TYPE = "type"
    FROM = "from"
    DESTINATION = "destination"
    CONTENT = "content"
    NODE_ID = "node_id"
    MODELS = "models"
    DATASETS = "datasets"
    NODES = "nodes"
    STATUS = "status"
    SUCCESS = "success"
    ERROR = "error"
    CPU = "cpu"
    MEM_USAGE = "mem_usage"


class WORKER_PROPERTIES:
    HEALTH_CHECK_INTERVAL = 15
    ONLINE = "online"
    BUSY = "busy"
    OFFLINE = "offline"
    PING_THRESHOLD = 100



from .. import http
from ..events import SocketHandler

from flask import Response
import json

socket_handler = SocketHandler()


@http.route("/models", methods=["GET"])
def get_models():
    try:
        response = {}
        models = []
        for node_id, node in socket_handler.nodes:
            models.append(node.hosted_models)

        response["models"] = models
        response_body = json.dumps(response)
        return Response(response_body, status=200, mimetype="application/json")
    except Exception:
        return Response({}, status=500, mimetype="application/json")



import time
import json
import requests
import re
import asyncio

from ..codes import MSG_FIELD, GRID_EVENTS, NODE_EVENTS, WORKER_PROPERTIES
from ..utils.wrappers import threaded


class Worker(object):
    def __init__(self, id: str, socket):
        self._id = id
        self._socket = socket
        self._ping = 0
        self._status = WORKER_PROPERTIES.ONLINE
        self.connected_nodes = {}
        self.hosted_models = {}
        self.hosted_datasets = {}
        self.cpu_percent = 0
        self.mem_usage = 0

    @property
    def status(self):
        if not self._socket:
            return WORKER_PROPERTIES.OFFLINE
        elif self._ping < WORKER_PROPERTIES.PING_THRESHOLD:
            return WORKER_PROPERTIES.ONLINE
        else:
            return WORKER_PROPERTIES.BUSY

    @property
    def address(self):
        if self._socket:
            addr = self._socket.environ["REMOTE_ADDR"]
            return re.sub("[:f]", "", addr)

    @property
    def location(self):
        if self.address:
            url = "http://ip-api.com/json/{}".format(self.address)
            r = requests.get(url)
            result = json.loads(r.text)
            if result["status"] == "success":
                return {
                    "region": result["regionName"],
                    "country": result["country"],
                    "city": result["city"],
                }
            else:
                return {}

    def send(self, message):
        self._socket.send(message)

    # Run it in a different thread
    @threaded
    def monitor(self):
        while self._socket:
            self.__begin = time.time()
            self._socket.send(json.dumps({MSG_FIELD.TYPE: NODE_EVENTS.MONITOR}))
            time.sleep(WORKER_PROPERTIES.HEALTH_CHECK_INTERVAL)

    def update_node_infos(self, message):
        if self.__begin:
            end = time.time()
            self._ping = (end - self.__begin) * 1000
            self.connected_nodes = message[MSG_FIELD.NODES]
            self.hosted_models = message[MSG_FIELD.MODELS]
            self.hosted_datasets = message[MSG_FIELD.DATASETS]
            self.cpu_percent = message[MSG_FIELD.CPU]
            self.mem_usage = message[MSG_FIELD.MEM_USAGE]



import queue

from ..workers import Worker
from ..codes import WORKER_PROPERTIES


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SocketHandler(metaclass=Singleton):
    """ Socket Handler is a sigleton class used to handle/manage websocket connections. """

    def __init__(self):
        self.connections = {}

    def new_connection(self, workerId: str, socket):
        """ Create a mapping structure to establish a bond between a workerId and a socket descriptor.
            Args:
                workerId: Uuid string used to identify workers.
                socket: Socket decriptor that will be used to send/receive messages from this client.
        """
        if workerId not in self.connections:
            self.connections[workerId] = Worker(workerId, socket)
        else:
            worker = self.connections[workerId]
            if worker.status == WORKER_PROPERTIES.OFFLINE:
                worker._socket = socket
        return self.connections[workerId]

    def send_msg(self, workerId: str, message: str):
        """ Find the socket descriptor mapped by workerId and send them a message.

            Args:
                workerId: Uiid used to identify and map workers.
                message: Message that will be send.
        """
        socket = self.connections.get(workerId, None)
        if socket:
            socket.send(message)

    def get(self, query):
        if isinstance(query, str):
            return self.connections.get(query, None)
        else:
            return self.__retrieve_worker_by_socket(query)

    def remove(self, socket) -> str:
        """ Remove a socket descriptor from mapping structure. It will be used when the socket connection is closed.

            Args:
                socket: socket descriptor used to send/receive messages.
            Returns:
                workerId: Worker id linked to that connection.
        """
        worker = self.__retrieve_worker_by_socket(socket)
        if worker:
            self.connections[worker._id]._socket = None
            self.connections[worker._id].connected_nodes = []
            return worker._id

    def __retrieve_worker_by_socket(self, socket):
        for worker_id, worker in self.connections.items():
            if worker._socket == socket:
                return self.connections[worker_id]

    @property
    def nodes(self) -> list:
        return list(self.connections.items())

    def __len__(self) -> int:
        """ Number of connections handled by this server.

            Returns:
                length : number of connections handled by this server.
        """
        return len(self.connections)



import os
from setuptools import find_packages, setup


def get_requirements(req_file):
    """Read requirements file and return packages and git repos separately"""
    requirements = []
    dependency_links = []
    lines = read(req_file).split("\n")
    for line in lines:
        if line.startswith("git+"):
            dependency_links.append(line)
        else:
            requirements.append(line)
    return requirements, dependency_links


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


core_reqs, core_dependency_links = get_requirements("requirements.txt")


setup(
    name="openmined.gridnetwork",
    version="0.1.0",
    author="OpenMined",
    author_email="contact@openmined.org",
    description="A network router used by the PyGrid platform.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/OpenMined/GridNetwork",
    maintainer="Benardi Nunes",
    maintainer_email="benardinunes@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=core_reqs,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={'console_scripts': ['raise_grid=gridnetwork:raise_grid']},
    python_requires=">=3.5, <4",
    project_urls={
        "Bug Reports": "https://github.com/OpenMined/GridNetwork/issues",
        "Funding": "https://opencollective.com/openmined",
        "Source": "https://github.com/OpenMined/GridNetwork",
    },
)



import threading
import json
import time

from .socket_handler import SocketHandler
from ..codes import MSG_FIELD

socket_handler = SocketHandler()


def update_node(message, socket):
    try:
        worker = socket_handler.get(socket)
        worker.update_node_infos(message)
    except Exception:
        pass


def register_node(message, socket):
    try:
        time.sleep(1)
        node_id = message[MSG_FIELD.NODE_ID]
        worker = socket_handler.new_connection(node_id, socket)
        t = threading.Thread(target=worker.monitor)
        t.start()
        return {"status": "success!"}
    except Exception:
        pass


def forward(message, socket):
    try:
        time.sleep(1)
        dest = message[MSG_FIELD.DESTINATION]
        worker = socket_handler.get(dest)
        if worker:
            content = message[MSG_FIELD.CONTENT]
            worker.send(json.dumps(content))
    except Exception:
        pass
